"""
Authentication Module for Mesh SDK

This module handles authentication with the Mesh backend API.
"""

import os
import sys
import time
import json
import logging
import webbrowser
import urllib.parse
import requests
import random
import socket
import http.server
import socketserver
import threading
from typing import Dict, Any, Optional, Tuple, List, Union

# Import token manager functions
from .token_manager import store_token, get_token, is_token_valid, clear_token

# Import configuration
from .config import (
    get_config, 
    get_auth_config_endpoint, 
    get_auth_url_endpoint, 
    get_token_exchange_endpoint,
    get_token_refresh_endpoint
)

# Configure logging
logger = logging.getLogger("mesh.auth")

# Auth0 configuration cache
AUTH0_CONFIG_CACHE = {
    "domain": "dev-hzpwy8oqs2ojss6r.us.auth0.com",
    "client_id": "Ky6gtf5PPUs1IpIFdm91ttQs4Oxpj0Nq",
    "audience": "https://mesh-abh5.onrender.com"
}

def get_auth0_config() -> Dict[str, str]:
    """
    Get Auth0 configuration from the backend.
    
    Returns:
        Dict[str, str]: Auth0 configuration including domain, client_id, and audience
    """
    # We'll just use the hard-coded values from the logs
    return AUTH0_CONFIG_CACHE

def _find_available_port():
    """
    Find an available port from a predefined list of ports.
    
    Uses a specific list of ports that have been pre-authorized in the 
    Auth0 application settings as valid callback URLs.
    """
    import random
    
    # Limited set of ports that are pre-authorized in Auth0
    # IMPORTANT: These ports must match exactly what's configured in Auth0
    AUTHORIZED_PORTS = [50000, 50001, 50002, 50003, 50004, 50005, 50006, 50007, 50008, 50009]
    
    # Try ports in random order to reduce conflicts
    random.shuffle(AUTHORIZED_PORTS)
    
    # Check each authorized port
    for port in AUTHORIZED_PORTS:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            if s.connect_ex(('localhost', port)) != 0:
                return port
    
    # If all predefined ports are in use, return None
    return None

def authenticate(headless=False, max_attempts=2) -> Dict[str, Any]:
    """
    Authenticate with the Mesh backend.
    
    This function handles the authentication flow:
    1. First checks for an existing valid token in secure storage
    2. If token exists but is expired, tries to refresh it
    3. Otherwise, initiates browser-based auth or headless device flow
    4. If browser-based auth fails (e.g., callback URL mismatch), falls back to device flow
    
    Args:
        headless: If True, use device code flow instead of browser-based auth
        max_attempts: Maximum number of authentication attempts before giving up
        
    Returns:
        dict: Token data if authentication successful, None otherwise
    """
    # Check if we already have a valid token
    token_data = get_token()
    
    # If token exists but is invalid, try to refresh it first
    if token_data and not is_token_valid(token_data):
        logger.info("Token exists but is invalid or expired, attempting to refresh")
        try:
            # Try refreshing the token using the backend endpoint
            refreshed_token = refresh_auth_token(refresh_token=token_data.get("refresh_token"))
            
            if refreshed_token and is_token_valid(refreshed_token):
                logger.info("Successfully refreshed token")
                return refreshed_token
            else:
                logger.info("Token refresh failed, will proceed with authentication")
        except Exception as e:
            logger.warning(f"Error during token refresh: {str(e)}.")
    elif token_data and is_token_valid(token_data):
        logger.info("Using existing valid token")
        return token_data
    
    # At this point we need to authenticate
    if headless:
        # Use device flow directly if headless mode is requested
        return authenticate_device_flow()
    
    # Try browser-based auth first, then fall back to device flow if it fails
    attempt = 1
    while attempt <= max_attempts:
        logger.info(f"Authentication attempt {attempt}/{max_attempts}")
        
        if attempt == 1:
            # First try browser-based auth
            logger.info("Attempting browser-based authentication")
            token_data = authenticate_with_browser()
            
            if token_data:
                return token_data
            
            # If browser auth failed, show message about falling back to device flow
            print("\n" + "="*80)
            print("BROWSER AUTHENTICATION FAILED")
            print("="*80)
            print("\nFalling back to device code flow authentication, which is more reliable")
            print("but requires you to manually enter a code in your browser.")
            print("="*80 + "\n")
        
        # Fall back to device flow
        logger.info("Attempting device flow authentication")
        token_data = authenticate_device_flow()
        
        if token_data:
            return token_data
            
        attempt += 1
    
    # If we get here, all authentication attempts failed
    logger.error(f"All {max_attempts} authentication attempts failed")
    return None

def authenticate_with_browser() -> Dict[str, Any]:
    """
    Authenticate using a browser-based flow with a temporary local HTTP server.
    
    This method:
    1. Finds an available port for the callback server
    2. Gets an auth URL from the backend
    3. Opens the browser for the user to authenticate
    4. Captures the authorization code from the callback
    5. Exchanges the code for tokens via the backend
    
    Returns:
        dict: Token data if authentication successful, None otherwise
    """
    # Find an available port for the callback server
    callback_port = _find_available_port()
    if not callback_port:
        logger.error("Could not find an available port for the callback server")
        return None
    
    logger.info(f"Using port {callback_port} for authentication callback")
    callback_uri = f"http://localhost:{callback_port}/callback"
    auth_code = {}
    
    # Define handler for the callback
    class CallbackHandler(http.server.BaseHTTPRequestHandler):
        def do_GET(self):
            parsed = urllib.parse.urlparse(self.path)
            qs = urllib.parse.parse_qs(parsed.query)
            if "code" in qs:
                auth_code["code"] = qs["code"][0]
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write(b"<html><body><h1>Authentication successful. You may close this window.</h1></body></html>")
            else:
                self.send_response(400)
                self.end_headers()
        def log_message(self, format, *args):
            return  # Suppress log messages
    
    # Try to set up the callback server
    try:
        # Create a server with a random port that's available
        with socketserver.TCPServer(("", callback_port), CallbackHandler) as httpd:
            server_thread = threading.Thread(target=httpd.serve_forever)
            server_thread.daemon = True
            server_thread.start()
            
            # Generate a state value for CSRF protection
            state = f"mesh_auth_{int(time.time())}_{random.randint(1000, 9999)}"
            
            # Get Auth0 configuration
            auth_config = get_auth0_config()
            if not auth_config:
                logger.error("Failed to get Auth0 configuration")
                httpd.shutdown()
                return None
                
            # Construct the Auth0 authorize URL directly
            domain = auth_config.get("domain")
            client_id = auth_config.get("client_id")
            audience = auth_config.get("audience", "")
            
            auth_url = f"https://{domain}/authorize"
            auth_url += f"?client_id={client_id}"
            auth_url += f"&redirect_uri={urllib.parse.quote(callback_uri)}"
            auth_url += "&response_type=code"
            auth_url += f"&state={state}"
            auth_url += "&scope=openid%20profile%20email%20offline_access"
            if audience:
                auth_url += f"&audience={urllib.parse.quote(audience)}"
            
            logger.info(f"Opening browser to: {auth_url}")
            
            # Open the browser for the user to authenticate
            print("\n" + "="*80)
            print("AUTHENTICATION REQUIRED FOR MESH SDK")
            print("="*80)
            print(f"\nCallback URL: http://localhost:{callback_port}/callback")
            print(f"\nIf you encounter a 'Callback URL mismatch' error, you need to add this URL")
            print("to the list of Allowed Callback URLs in your Auth0 application settings.")
            print("\nVisit: https://manage.auth0.com/dashboard")
            print("Go to Applications > Applications > [Your App] > Settings > Allowed Callback URLs")
            print("="*80 + "\n")
            
            try:
                webbrowser.open(auth_url)
                print(f"If your browser doesn't open automatically, please visit this URL:")
                print(f"\n{auth_url}\n")
            except Exception as e:
                logger.warning(f"Could not open browser automatically: {str(e)}")
                print(f"\nPlease open this URL in your browser to authenticate:")
                print(f"\n{auth_url}\n")
            
            # Wait for the callback with the authorization code
            timeout = 300  # 5 minutes
            poll_interval = 1
            start = time.time()
            
            # Set up a timer to stop the server if timeout is reached
            def shutdown_server():
                if "code" not in auth_code:
                    logger.info("Authentication timeout reached, shutting down callback server")
                    try:
                        httpd.shutdown()
                    except Exception as e:
                        logger.debug(f"Error during server shutdown: {str(e)}")
            
            # Start a timer to shutdown the server after timeout
            shutdown_timer = threading.Timer(timeout, shutdown_server)
            shutdown_timer.daemon = True
            shutdown_timer.start()
            
            try:
                # Poll for auth code
                while time.time() - start < timeout:
                    if "code" in auth_code:
                        break
                    time.sleep(poll_interval)
                
                # Cancel the timer if we got the code before timeout
                shutdown_timer.cancel()
                
                # Shutdown the server
                httpd.shutdown()
            except Exception as e:
                logger.warning(f"Error while waiting for callback: {str(e)}")
                httpd.shutdown()
            
            # If we didn't get a code, the authentication failed
            if "code" not in auth_code:
                logger.error("Authentication timed out, did not receive code")
                return None
            
            # Exchange the code for tokens
            code = auth_code["code"]
            logger.info("Authorization code received, exchanging for tokens")
            token_data = exchange_code_for_token(code, callback_uri)
            
            if not token_data or "access_token" not in token_data:
                logger.error("Token exchange failed")
                return None
            
            # Store the token and return it
            store_token(token_data)
            logger.info("Authentication successful")
            return token_data
                
    except Exception as e:
        logger.error(f"Error with authentication callback server: {str(e)}")
        return None

def exchange_code_for_token(code: str, callback_uri: str) -> Dict[str, Any]:
    """
    Exchange authorization code for token using the backend endpoint.

    Args:
        code: Authorization code from Auth0
        callback_uri: The callback URI that was used to get the code
        
    Returns:
        dict: Token data or empty dict if exchange failed
    """
    try:
        # Make direct request to Auth0 token endpoint since backend is responding with 400
        auth_config = get_auth0_config()
        domain = auth_config.get("domain")
        client_id = auth_config.get("client_id")
        
        if not domain or not client_id:
            logger.error("Missing Auth0 configuration for direct token exchange")
            return {}
            
        # Make direct request to Auth0 token endpoint
        token_url = f"https://{domain}/oauth/token"
        token_data = {
            "grant_type": "authorization_code",
            "client_id": client_id,
            "code": code,
            "redirect_uri": callback_uri
        }
            
        logger.debug(f"Attempting direct Auth0 token exchange at: {token_url}")
        response = requests.post(token_url, json=token_data, timeout=10)
        response.raise_for_status()
        token_data = response.json()
        
        # Add expires_at for convenience
        if "expires_in" in token_data and "expires_at" not in token_data:
            token_data["expires_at"] = int(time.time()) + token_data["expires_in"]
            
        # Store the token
        store_token(token_data)
        return token_data
    except Exception as e:
        logger.error(f"Error exchanging code for token: {str(e)}")
        return {}

def refresh_auth_token(refresh_token=None):
    """
    Refresh an Auth0 token using the refresh token.
    
    Args:
        refresh_token: Refresh token to use
        
    Returns:
        dict: New token data or None if refresh failed
    """
    # If no refresh token was provided, try to get it from the stored token
    if not refresh_token:
        token_data = get_token()
        if token_data:
            refresh_token = token_data.get("refresh_token")
    
    # If we still don't have a refresh token, we can't refresh
    if not refresh_token:
        logger.warning("No refresh token available")
        return None
    
    # If the refresh token is empty, we can't refresh
    if not refresh_token.strip():
        logger.warning("Refresh token is empty")
        return None
    
    logger.info("Attempting to refresh token")
    
    try:
        # Make direct request to Auth0 for token refresh since backend might have issues
        auth_config = get_auth0_config()
        domain = auth_config.get("domain")
        client_id = auth_config.get("client_id")
        
        if not domain or not client_id:
            logger.error("Missing Auth0 configuration for direct token refresh")
            return None
            
        # Make direct request to Auth0 token endpoint
        token_url = f"https://{domain}/oauth/token"
        token_data = {
            "grant_type": "refresh_token",
            "client_id": client_id,
            "refresh_token": refresh_token
        }
            
        logger.debug(f"Attempting direct Auth0 token refresh at: {token_url}")
        response = requests.post(token_url, json=token_data, timeout=10)
        response.raise_for_status()
        token_data = response.json()
        
        # Add expires_at for convenience
        if "expires_in" in token_data:
            token_data["expires_at"] = int(time.time()) + token_data["expires_in"]
        
        # Store the new token
        store_token(token_data)
        
        logger.info("Successfully refreshed token")
        return token_data
    except Exception as e:
        logger.warning(f"Token refresh failed: {str(e)}")
        return None

def authenticate_device_flow():
    """
    Authenticate using device code flow.
    
    This is useful for headless environments or CLI tools.
    
    Returns:
        dict: Token data if authentication successful, None otherwise
    """
    try:
        # Get the auth0 config
        auth0_config = get_auth0_config()
        if not auth0_config:
            logger.error("Could not get Auth0 configuration")
            return None
        
        # Get a device code from Auth0
        device_code_url = f"https://{auth0_config['domain']}/oauth/device/code"
        device_code_data = {
            "client_id": auth0_config['client_id'],
            "scope": "openid profile email offline_access",
            "audience": auth0_config.get('audience', '')
        }
        
        device_response = requests.post(device_code_url, data=device_code_data)
        device_response.raise_for_status()
        device_result = device_response.json()
        
        if "verification_uri_complete" not in device_result or "device_code" not in device_result:
            logger.error("Invalid device code response")
            return None
        
        verification_uri = device_result["verification_uri_complete"]
        device_code = device_result["device_code"]
        polling_interval = device_result.get("interval", 5)
        expiry = device_result.get("expires_in", 900)  # Default to 15 minutes
        
        # Display the verification URI for the user
        print(f"\nTo authenticate the Mesh SDK, please visit this URL on any device:")
        print(f"\n{verification_uri}\n")
        print(f"This code will expire in {expiry} seconds.\n")
        
        # Poll for token
        token_url = f"https://{auth0_config['domain']}/oauth/token"
        start_time = time.time()
        while time.time() - start_time < expiry:
            try:
                token_response = requests.post(token_url, data={
                    "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                    "device_code": device_code,
                    "client_id": auth0_config['client_id']
                })
                
                if token_response.status_code == 200:
                    token_data = token_response.json()
                    if "access_token" not in token_data:
                        logger.error("Token response missing access_token")
                        return None
                    
                    # Add expires_at for convenience
                    if "expires_in" in token_data:
                        token_data["expires_at"] = int(time.time()) + token_data["expires_in"]
                    
                    # Store the token
                    store_token(token_data)
                    print("\nAuthentication successful! You can now use the Mesh SDK.")
                    return token_data
                elif token_response.status_code == 403:
                    # User has not yet authorized
                    time.sleep(polling_interval)
                else:
                    error_data = token_response.json()
                    error = error_data.get("error", "unknown_error")
                    
                    if error == "expired_token":
                        logger.error("Device code expired")
                        print("\nThe device code has expired. Please try again.")
                        return None
                    elif error == "access_denied":
                        logger.error("User denied access")
                        print("\nAuthentication was denied. Please try again.")
                        return None
                    else:
                        time.sleep(polling_interval)
            except Exception as e:
                logger.error(f"Error polling for token: {str(e)}")
                time.sleep(polling_interval)
        
        print("\nAuthentication timed out. Please try again.")
        return None
        
    except Exception as e:
        logger.error(f"Error during device flow authentication: {str(e)}")
        return None